@@ -63,6 +63,9 @@ struct ClaspConfig::Impl {
 			}
 			ptr()->prepare(ctx);
 		}
+		void unfreeze(SharedContext& ctx) {
+			ptr()->unfreeze(ctx);
+		}
 		void destroy() { if (test_bit(cfg, AcquireBit)) { delete ptr(); } }
 		Configurator* ptr() const {
 			static const uint64 ptrMask = ~(bit_mask<uint64>(OnceBit) | bit_mask<uint64>(AcquireBit));
@@ -78,6 +81,7 @@ struct ClaspConfig::Impl {
 	void prepare(SharedContext& ctx);
 	bool addPost(Solver& s, const SolverParams& opts);
 	void add(Configurator* c, Ownership_t::Type t, bool once) { pp.push_back(ConfiguratorProxy(c, t, once)); }
+	void unfreeze(SharedContext& ctx);
 	PPVec   pp;
 	uint64  acycSet;
 #if CLASP_HAS_THREADS
@@ -126,6 +130,11 @@ bool ClaspConfig::Impl::addPost(Solver& s, const SolverParams& opts) {
 	return true;
 #undef LOCKED
 }
+void ClaspConfig::Impl::unfreeze(SharedContext& ctx) {
+	for (PPVec::iterator it = pp.begin(), end = pp.end(); it != end; ++it) {
+		it->unfreeze(ctx);
+	}
+}
 
 ClaspConfig::ClaspConfig() : tester_(0), impl_(new Impl()) {}
 ClaspConfig::~ClaspConfig() {
@@ -182,8 +191,13 @@ bool ClaspConfig::addPost(Solver& s) const {
 	return impl_->addPost(s, solver(s.id())) && BasicSatConfig::addPost(s);
 }
 
+void ClaspConfig::unfreeze(SharedContext& ctx) {
+	impl_->unfreeze(ctx);
+}
+
 ClaspConfig::Configurator::~Configurator() {}
 void ClaspConfig::Configurator::prepare(SharedContext&) {}
+void ClaspConfig::Configurator::unfreeze(SharedContext&){}
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspFacade::SolveStrategy
 /////////////////////////////////////////////////////////////////////////////////////////
@@ -555,6 +569,7 @@ double _getResult(const SolveResult* r) { return static_cast<double>(r->operator
 double _getSignal(const SolveResult* r) { return static_cast<double>(r->signal); }
 double _getExhausted(const SolveResult* r) { return static_cast<double>(r->exhausted()); }
 }
+
 struct ClaspFacade::Statistics {
 	Statistics(ClaspFacade& f) : self_(&f), tester_(0), level_(0), clingo_(0) {}
 	~Statistics() { delete clingo_; delete solvers_.multi; }
@@ -564,22 +579,22 @@ struct ClaspFacade::Statistics {
 	void addTo(StatsMap& solving, StatsMap* accu) const;
 	void accept(StatsVisitor& out, bool final) const;
 	bool incremental() const { return self_->incremental(); }
-	Potassco::AbstractStatistics* getClingo();
 	typedef StatsVec<SolverStats>        SolverVec;
 	typedef SingleOwnerPtr<Asp::LpStats> LpStatsPtr;
 	typedef PrgDepGraph::NonHcfStats     TesterStats;
-	ClaspFacade*  self_;
-	LpStatsPtr    lp_;      // level 0 and asp
-	SolverStats   solvers_; // level 0
-	SolverVec     solver_;  // level > 1
-	SolverVec     accu_;    // level > 1 and incremental
-	TesterStats*  tester_;  // level > 0 and nonhcfs
-	uint32        level_;   // active stats level
+	ClaspFacade* self_;
+	LpStatsPtr   lp_;      // level 0 and asp
+	SolverStats  solvers_; // level 0
+	SolverVec    solver_;  // level > 1
+	SolverVec    accu_;    // level > 1 and incremental
+	TesterStats* tester_;  // level > 0 and nonhcfs
+	uint32       level_;   // active stats level
 	// For clingo stats interface
 	class ClingoView : public ClaspStatistics {
 	public:
 		explicit ClingoView(const ClaspFacade& f);
 		void update(const Statistics& s);
+		Key_t user(bool final) const;
 	private:
 		struct StepStats {
 			SummaryStats times;
@@ -593,27 +608,26 @@ struct ClaspFacade::Statistics {
 				summary.add("models", models.toStats());
 			}
 		};
-		StatsMap keys_;
-		StatsMap problem_;
-		StatsMap solving_;
+		StatsMap*   keys_;
+		StatsMap    problem_;
+		StatsMap    solving_;
 		struct Summary : StatsMap { StepStats step; } summary_;
 		struct Accu    : StatsMap { StepStats step; StatsMap solving_; };
 		typedef SingleOwnerPtr<Accu> AccuPtr;
 		AccuPtr accu_;
-	}* clingo_; // new clingo stats interface;
+	}* clingo_; // new clingo stats interface
+	ClingoView* getClingo();
 };
 void ClaspFacade::Statistics::initLevel(uint32 level) {
 	if (level_ < level) {
 		if (incremental() && !solvers_.multi) { solvers_.multi = new SolverStats(); }
 		level_ = level;
 	}
-	if (self_->isAsp() && !lp_.get()) {
-		lp_ = new Asp::LpStats();
-	}
 	if (self_->ctx.sccGraph.get() && self_->ctx.sccGraph->numNonHcfs() && !tester_) {
 		tester_ = self_->ctx.sccGraph->nonHcfStats();
 	}
 }
+
 void ClaspFacade::Statistics::start(uint32 level) {
 	// cleanup previous state
 	solvers_.reset();
@@ -658,6 +672,9 @@ void ClaspFacade::Statistics::accept(StatsVisitor& out, bool final) const {
 		const SolverVec& solver = final ? accu_ : solver_;
 		const uint32 nThreads = final ? (uint32)accu_.size() : self_->ctx.concurrency();
 		const uint32 nSolver  = (uint32)solver.size();
+		if (const AbstractStatistics::Key_t userKey = clingo_ ? clingo_->user(final) : 0) {
+			out.visitExternalStats(clingo_->getObject(userKey));
+		}
 		if (nThreads > 1 && nSolver > 1 && out.visitThreads(StatsVisitor::Enter)) {
 			for (uint32 i = 0, end = std::min(nSolver, nThreads); i != end; ++i) {
 				out.visitThread(i, *solver[i]);
@@ -671,7 +688,7 @@ void ClaspFacade::Statistics::accept(StatsVisitor& out, bool final) const {
 		out.visitTester(StatsVisitor::Leave);
 	}
 }
-Potassco::AbstractStatistics* ClaspFacade::Statistics::getClingo() {
+ClaspFacade::Statistics::ClingoView* ClaspFacade::Statistics::getClingo() {
 	if (!clingo_) {
 		clingo_ = new ClingoView(*this->self_);
 		clingo_->update(*this);
@@ -679,6 +696,7 @@ Potassco::AbstractStatistics* ClaspFacade::Statistics::getClingo() {
 	return clingo_;
 }
 ClaspFacade::Statistics::ClingoView::ClingoView(const ClaspFacade& f) {
+	keys_ = makeRoot();
 	summary_.add("call"       , StatisticObject::value(&f.step_.step));
 	summary_.add("result"     , StatisticObject::value<SolveResult, _getResult>(&f.step_.result));
 	summary_.add("signal"     , StatisticObject::value<SolveResult, _getSignal>(&f.step_.result));
@@ -693,17 +711,22 @@ ClaspFacade::Statistics::ClingoView::ClingoView(const ClaspFacade& f) {
 		if (f.incremental()) { problem_.add("lpStep", StatisticObject::map(f.step_.lpStep())); }
 	}
 	problem_.add("generator", StatisticObject::map(&f.ctx.stats()));
-	keys_.add("problem", problem_.toStats());
-	keys_.add("solving", solving_.toStats());
-	keys_.add("summary", summary_.toStats());
+	keys_->add("problem", problem_.toStats());
+	keys_->add("solving", solving_.toStats());
+	keys_->add("summary", summary_.toStats());
+
 	if (f.incremental()) {
 		accu_ = new Accu();
 		accu_->step.bind(*f.accu_.get());
 	}
-	setRoot(keys_.toStats());
+}
+Potassco::AbstractStatistics::Key_t ClaspFacade::Statistics::ClingoView::user(bool final) const {
+	Key_t key = 0;
+	find(root(), final ? "user_accu" : "user_step", &key);
+	return key;
 }
 void ClaspFacade::Statistics::ClingoView::update(const ClaspFacade::Statistics& stats) {
-	if (stats.level_ > 0 && accu_.get() && keys_.add("accu", accu_->toStats())) {
+	if (stats.level_ > 0 && accu_.get() && keys_->add("accu", accu_->toStats())) {
 		accu_->step.addTo(*accu_);
 		accu_->add("solving", accu_->solving_.toStats());
 	}
@@ -736,7 +759,7 @@ ProblemType ClaspFacade::detectProblemType(std::istream& str) {
 	return Clasp::detectProblemType(str);
 }
 const ClaspFacade::Summary&  ClaspFacade::summary(bool accu) const {
-	return accu && accu_.get() && accu_->step ? *accu_ : step_;
+	return accu && accu_.get() ? *accu_ : step_;
 }
 
 void ClaspFacade::discardProblem() {
@@ -814,6 +837,7 @@ Asp::LogicProgram& ClaspFacade::startAsp(ClaspConfig& config, bool enableUpdates
 	p->setOptions(config.asp);
 	p->setNonHcfConfiguration(config.testerConfig());
 	type_ = Problem_t::Asp;
+	stats_->lp_ = new Asp::LpStats();
 	if (enableUpdates) { enableProgramUpdates(); }
 	return *p;
 }
@@ -914,6 +938,7 @@ bool ClaspFacade::read() {
 	if (!p.more()) { p.reset(); }
 	return true;
 }
+
 void ClaspFacade::prepare(EnumMode enumMode) {
 	POTASSCO_REQUIRE(solve_.get() && !solving());
 	EnumOptions& en = config_->solve;
@@ -986,6 +1011,7 @@ void ClaspFacade::doUpdate(ProgramBuilder* p, bool updateConfig, void(*sigAct)(i
 		ctx.unfreeze();
 	}
 	solve_->reset();
+	config_->unfreeze(ctx);
 	int sig = sigAct == SIG_DFL ? 0 : solve_->qSig.exchange(0);
 	if (sig && sigAct != SIG_IGN) { sigAct(sig); }
 }
