@@ -568,14 +568,20 @@ Id_t LogicProgram::newCondition(const Potassco::LitSpan& cond) {
 	return static_cast<Id_t>(Clasp::Asp::falseId);
 }
 LogicProgram& LogicProgram::addOutput(const ConstString& str, const Potassco::LitSpan& cond) {
+	if (cond.size == 1) {
+		POTASSCO_REQUIRE(Potassco::atom(cond[0]) < bodyId, "Atom out of bounds");
+		return addOutput(str, Potassco::id(cond[0]));
+	}
 	if (!ctx()->output.filter(str)) {
-		POTASSCO_REQUIRE(cond.size != 1 || Potassco::atom(cond[0]) < bodyId, "Atom out of bounds");
-		show_.push_back(ShowPair(cond.size == 1 ? Potassco::id(cond[0]) : newCondition(cond), str));
+		show_.push_back(ShowPair(newCondition(cond), str));
 	}
 	return *this;
 }
 LogicProgram& LogicProgram::addOutput(const ConstString& str, Id_t id) {
 	if (!ctx()->output.filter(str) && id != falseId) {
+		if (Potassco::atom(id) < bodyId) {
+			resize(Potassco::atom(id));
+		}
 		show_.push_back(ShowPair(id, str));
 	}
 	return *this;
@@ -798,7 +804,7 @@ Literal LogicProgram::getLiteral(Id_t id, MapLit_t m) const {
 				Var v = isNew(id)
 					? incData_->steps.back().second
 					: std::lower_bound(incData_->steps.begin(), incData_->steps.end(), Incremental::StepTrue(nId, 0))->second;
-				out = posLit(v);
+				out = Literal(v, out.sign());
 			}
 		}
 	}
@@ -1034,6 +1040,7 @@ void LogicProgram::prepareExternals() {
 			*j++ = encodeExternal(id, value);
 		}
 	}
+	external.erase(j, external.end());
 	for (VarVec::const_iterator it = external.begin(), end = external.end(); it != end; ++it) {
 		atomState_.clearRule(decodeExternal(*it).first);
 	}
@@ -1137,6 +1144,7 @@ void LogicProgram::prepareProgram(bool checkSccs) {
 	finalizeDisjunctions(p, sccs);
 	prepareComponents();
 	prepareOutputTable();
+	freezeAssumptions();
 	if (incData_ && options().distTrue) {
 		for (Var a = startAtom(), end = startAuxAtom(); a != end; ++a) {
 			if (isSentinel(getRootAtom(a)->literal())) {
@@ -1401,6 +1409,17 @@ void LogicProgram::prepareOutputTable() {
 		}
 	}
 }
+
+// Make assumptions/externals exempt from sat-preprocessing
+void LogicProgram::freezeAssumptions() {
+	for (VarVec::const_iterator it = frozen_.begin(), end = frozen_.end(); it != end; ++it) {
+		ctx()->setFrozen(getRootAtom(*it)->var(), true);
+	}
+	for (Potassco::LitVec::const_iterator it = auxData_->assume.begin(), end = auxData_->assume.end(); it != end; ++it) {
+		ctx()->setFrozen(getLiteral(Potassco::id(*it)).var(), true);
+	}
+}
+
 // add (completion) nogoods
 bool LogicProgram::addConstraints() {
 	ClauseCreator gc(ctx()->master());
@@ -1434,7 +1453,7 @@ bool LogicProgram::addConstraints() {
 		if (!toConstraint(getAtom(*it), *this, gc)) { return false; }
 	}
 	// add atoms from this step
-	const bool freezeAll = incData_ && ctx()->satPrepro.get() != 0;
+	const bool freezeAll = incData_ != 0;
 	const uint32 hiAtom  = startAuxAtom();
 	uint32 id = startAtom();
 	for (AtomList::const_iterator it = atoms_.begin()+startAtom(), end = atoms_.end(); it != end; ++it, ++id) {
@@ -1661,7 +1680,9 @@ bool LogicProgram::simplifySum(Head_t ht, const Potassco::AtomSpan& head, const
 				meta.hash += hashLit(p);
 			}
 			else { // Merge duplicate lits
-				w = (std::find_if(out.wlits_begin(), out.wlits_end(), IsLit(Potassco::lit(*it)))->weight += w);
+				Potassco::WeightLit_t* pos = std::find_if(out.wlits_begin(), out.wlits_end(), IsLit(toInt(p)));
+				POTASSCO_ASSERT(pos != out.wlits_end());
+				w = (pos->weight += w);
 				++dirty;
 			}
 			if (w > maxW) { maxW = w; }
@@ -1752,7 +1773,7 @@ Literal LogicProgram::getEqAtomLit(Literal lit, const BodyList& supports, Prepro
 	if (supports.empty() || lit == lit_false()) {
 		return lit_false();
 	}
-	else if (supports.size() == 1 && supports[0]->size() < 2) {
+	else if (supports.size() == 1 && supports[0]->size() < 2 && supports.back()->literal() == lit) {
 		return supports[0]->size() == 0 ? lit_true() : supports[0]->goal(0);
 	}
 	else if (p.getRootAtom(lit) != varMax) {
