@@ -356,6 +356,44 @@ TEST_CASE("Facade", "[facade]") {
 		REQUIRE(libclasp.summary().model()->isTrue(asp.getLiteral(1)));
 		REQUIRE(libclasp.summary().model()->isTrue(asp.getLiteral(2)));
 	}
+	SECTION("testUncoreUndoerAssumptions") {
+		config.solve.numModels = 0;
+		config.solve.optMode   = MinimizeMode_t::enumOpt;
+		config.addSolver(0).heuId = Heuristic_t::Domain;
+		SECTION("test oll") {
+			config.addSolver(0).opt.type = OptParams::type_usc;
+			config.addSolver(0).opt.algo = OptParams::usc_oll;
+		}
+		SECTION("test one") {
+			config.addSolver(0).opt.type = OptParams::type_usc;
+			config.addSolver(0).opt.algo = OptParams::usc_one;
+		}
+		SECTION("test k") {
+			config.addSolver(0).opt.type = OptParams::type_usc;
+			config.addSolver(0).opt.algo = OptParams::usc_k;
+		}
+		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
+		lpAdd(asp,
+			"{x1;x2;x3;x4;x5}.\n"
+			":- x1, x2, x3.\n"
+			":- x4, x5.\n"
+			":- x4, not x5.\n"
+			"x5 :- not x4.\n"
+			"#minimize{not x1, not x2}.\n"
+			"#heuristic x4. [1,true]"
+			"#assume{x3}.\n");
+		libclasp.prepare();
+		REQUIRE(libclasp.solve().sat());
+		REQUIRE(libclasp.summary().numOptimal == 2);
+		libclasp.update();
+		libclasp.ctx.addUnary(~asp.getLiteral(3));
+		libclasp.ctx.addUnary(asp.getLiteral(5));
+		libclasp.prepare();
+		REQUIRE(libclasp.solve().sat());
+		REQUIRE(libclasp.summary().costs()->at(0) == 0);
+		REQUIRE(libclasp.summary().numOptimal == 1);
+	}
+
 	SECTION("testUpdateConfig") {
 		config.solve.numModels = 0;
 		config.solve.enumMode  = EnumOptions::enum_auto;
@@ -943,6 +981,36 @@ static void getStatsKeys(const Potassco::AbstractStatistics& stats, Potassco::Ab
 	}
 }
 
+static void addExternalStats(Potassco::AbstractStatistics* us, Potassco::AbstractStatistics::Key_t userRoot) {
+	typedef Potassco::AbstractStatistics::Key_t Key_t;
+
+	Key_t general = us->add(userRoot, "deathCounter", Potassco::Statistics_t::Map);
+	REQUIRE(us->get(userRoot, "deathCounter") == general);
+	REQUIRE(us->type(general) == Potassco::Statistics_t::Map);
+	Key_t value   = us->add(general, "total", Potassco::Statistics_t::Value);
+	us->set(value, 42.0);
+	value = us->add(general, "chickens", Potassco::Statistics_t::Value);
+	us->set(value, 712.0);
+
+	Key_t array = us->add(general, "thread", Potassco::Statistics_t::Array);
+	REQUIRE(us->get(general, "thread") == array);
+	REQUIRE(us->type(array) == Potassco::Statistics_t::Array);
+	REQUIRE(us->size(array) == 0);
+	for (size_t t = 0; t != 4; ++t) {
+		Key_t a = us->push(array, Potassco::Statistics_t::Map);
+		value   = us->add(a, "total", Potassco::Statistics_t::Value);
+		us->set(value, 20*(t+1));
+		Key_t m = us->add(a, "Animals", Potassco::Statistics_t::Map);
+		value   = us->add(m, "chicken", Potassco::Statistics_t::Value);
+		us->set(value, 2*(t+1));
+		value   = us->add(m, "cows", Potassco::Statistics_t::Value);
+		us->set(value, 5*(t+1));
+		value   = us->add(a, "feeding cost", Potassco::Statistics_t::Value);
+		us->set(value, t+1);
+	}
+	REQUIRE(us->add(userRoot, "deathCounter", Potassco::Statistics_t::Map) == general);
+}
+
 TEST_CASE("Facade statistics", "[facade]") {
 	Clasp::ClaspFacade libclasp;
 	Clasp::ClaspConfig config;
@@ -956,7 +1024,10 @@ TEST_CASE("Facade statistics", "[facade]") {
 		typedef Potassco::AbstractStatistics::Key_t Key_t;
 		Key_t r = stats->root();
 		REQUIRE(stats->type(r) == Potassco::Statistics_t::Map);
+		REQUIRE(stats->writable(r) == true);
 		Key_t lp = stats->get(r, "problem.lp");
+		REQUIRE(stats->writable(lp) == false);
+
 		Key_t s = stats->get(r, "solving");
 		Key_t m = stats->get(r, "summary.models");
 		REQUIRE(stats->type(lp) == Potassco::Statistics_t::Map);
@@ -976,9 +1047,18 @@ TEST_CASE("Facade statistics", "[facade]") {
 		getStatsKeys(*stats, r, keys, "");
 		REQUIRE(!keys.empty());
 		for (std::vector<std::string>::const_iterator it = keys.begin(), end = keys.end(); it != end; ++it) {
-			REQUIRE(stats->type(stats->get(r, it->c_str())) == Potassco::Statistics_t::Value);
+			Key_t result;
+			REQUIRE(stats->find(r, it->c_str(), &result));
+			REQUIRE(result == stats->get(r, it->c_str()));
+			REQUIRE(stats->type(result) == Potassco::Statistics_t::Value);
 		}
 		REQUIRE(keys.size() == 237);
+
+		Key_t result;
+		REQUIRE(stats->find(r, "problem.lp", &result));
+		REQUIRE(result == lp);
+		REQUIRE(!stats->find(lp, "foo", 0));
+		REQUIRE(stats->find(lp, "rules", &result));
 	}
 	SECTION("testClingoStatsKeyIntegrity") {
 		config.addTesterConfig()->stats = 2;
@@ -1046,6 +1126,8 @@ TEST_CASE("Facade statistics", "[facade]") {
 		REQUIRE(stats->get(root, "problem") != root);
 		REQUIRE(stats->get(root, "summary") != root);
 		REQUIRE_THROWS_AS(stats->get(root, "solving.accu"), std::out_of_range);
+		Key_t solving = stats->get(root, "solving");
+		REQUIRE(stats->find(solving, "accu", 0) == false);
 	}
 	SECTION("testClingoStatsBug") {
 		config.stats = 0;
@@ -1070,6 +1152,86 @@ TEST_CASE("Facade statistics", "[facade]") {
 		REQUIRE(stats->size(costs) == 0);
 		REQUIRE_THROWS_AS(stats->value(minVal), std::logic_error);
 	}
+	SECTION("testWritableStats") {
+		ClaspStatistics stats;
+		typedef ClaspStatistics::Key_t Key_t;
+		StatsMap* rootMap = stats.makeRoot();
+		double v1 = 2.0;
+		rootMap->add("fixed", StatisticObject::value(&v1));
+
+		Key_t root = stats.root();
+		REQUIRE(stats.writable(root));
+		REQUIRE(stats.writable(stats.get(root, "fixed")) == false);
+
+		Key_t v2 = stats.add(root, "mutable", Potassco::Statistics_t::Value);
+		REQUIRE(stats.writable(v2));
+		stats.set(v2, 22.0);
+		REQUIRE(stats.value(v2) == 22.0);
+		Key_t found;
+		REQUIRE(stats.find(root, "mutable", &found));
+		REQUIRE(found == v2);
+
+		Key_t arr = stats.add(root, "array", Potassco::Statistics_t::Array);
+		REQUIRE(stats.type(arr) == Potassco::Statistics_t::Array);
+		REQUIRE(stats.writable(arr));
+		REQUIRE(stats.size(arr) == 0);
+
+		Key_t mapAtArr0 = stats.push(arr, Potassco::Statistics_t::Map);
+		REQUIRE(stats.type(mapAtArr0) == Potassco::Statistics_t::Map);
+		REQUIRE(stats.size(arr) == 1);
+	}
+	SECTION("testClingoUserStats") {
+		typedef Potassco::AbstractStatistics::Key_t Key_t;
+		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
+		lpAdd(asp, "{x1;x2;x3}. #minimize{x1, x2}.");
+		libclasp.prepare();
+		libclasp.solve();
+		Potassco::AbstractStatistics* stats = libclasp.getStats();
+
+		Key_t r = stats->root();
+		REQUIRE(stats->type(r) == Potassco::Statistics_t::Map);
+
+		Key_t u = stats->add(r, "user_step", Potassco::Statistics_t::Map);
+		addExternalStats(stats, u);
+
+		REQUIRE(stats->type(u) == Potassco::Statistics_t::Map);
+		Key_t user = stats->get(u, "deathCounter");
+		REQUIRE(stats->type(user) == Potassco::Statistics_t::Map);
+		REQUIRE(stats->value(stats->get(user, "total")) == double(42));
+		REQUIRE(stats->value(stats->get(user, "chickens")) == double(712));
+		Key_t array = stats->get(user, "thread");
+		REQUIRE(stats->type(array) == Potassco::Statistics_t::Array);
+		REQUIRE(stats->size(array) == 4);
+		for (size_t t = 0; t != 4; ++t) {
+			Key_t m1 = stats->at(array, t);
+			REQUIRE(stats->type(m1) == Potassco::Statistics_t::Map);
+			Key_t value = stats->get(m1, "total");
+			REQUIRE(stats->type(value) == Potassco::Statistics_t::Value);
+			REQUIRE(stats->value(value) == double(20*(t+1)));
+			Key_t m2 = stats->get(m1, "Animals");
+			REQUIRE(stats->type(m2) == Potassco::Statistics_t::Map);
+			value = stats->get(m2, "chicken");
+			REQUIRE(stats->value(value) == double(2*(t+1)));
+			value = stats->get(m2, "cows");
+			REQUIRE(stats->value(value) == double(5*(t+1)));
+			value = stats->get(m1, "feeding cost");
+			REQUIRE(stats->value(value) == double(t+1));
+		}
+		Key_t total;
+		REQUIRE(stats->find(r, "user_step.deathCounter.thread.1.total", &total));
+		REQUIRE(stats->type(total) == Potassco::Statistics_t::Value);
+		REQUIRE(stats->value(total) == 40.0);
+		REQUIRE(!stats->find(r, "user_step.deathCounter.thread.5.total", 0));
+
+		std::vector<std::string> keys;
+		getStatsKeys(*stats, r, keys, "");
+		REQUIRE(!keys.empty());
+		for (std::vector<std::string>::const_iterator it = keys.begin(), end = keys.end(); it != end; ++it) {
+			REQUIRE(stats->find(r, it->c_str(), 0));
+			REQUIRE(stats->type(stats->get(r, it->c_str())) == Potassco::Statistics_t::Value);
+		}
+		REQUIRE(keys.size() == 255);
+	}
 }
 namespace {
 class MyProp : public Potassco::AbstractPropagator {
@@ -1522,15 +1684,52 @@ TEST_CASE("Clingo propagator", "[facade][propagator]") {
 			libclasp.prepare();
 			REQUIRE(libclasp.solve().unsat());
 		}
+
+		SECTION("testRelocationBug") {
+			class Prop : public Potassco::AbstractPropagator {
+			public:
+				Prop() {}
+				virtual void propagate(Potassco::AbstractSolver& s, const ChangeList& changes) {
+					Potassco::LitVec cmp(begin(changes), end(changes));
+					Potassco::LitVec clause;
+					clause.assign(1, 0);
+					for (uint32 i = 1; i <= s.assignment().level(); ++i) {
+						clause.push_back(-s.assignment().decision(i));
+					}
+					for (Potassco::Lit_t lit = 1; s.assignment().hasLit(lit); ++lit) {
+						if (s.assignment().value(lit) == Potassco::Value_t::Free) {
+							clause[0] = lit;
+							s.addClause(Potassco::toSpan(clause));
+							s.propagate();
+						}
+					}
+					REQUIRE(std::memcmp(&cmp[0], changes.first, changes.size * sizeof(Potassco::Lit_t)) == 0);
+				}
+				virtual void undo(const Potassco::AbstractSolver&, const ChangeList&) {}
+				virtual void check(Potassco::AbstractSolver&) {}
+			} prop;
+			MyInit pp(prop);
+			config.addConfigurator(&pp);
+			Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
+			lpAdd(asp, "{x1;x2;x3;x4;x5;x6;x7;x8;x9;x10;x11;x12;x13;x14;x15;x16}.");
+			asp.endProgram();
+			for (Var v = 1; v <= libclasp.ctx.numVars(); ++v) {
+				pp.addWatch(posLit(v));
+				pp.addWatch(negLit(v));
+			}
+			libclasp.prepare();
+			REQUIRE(libclasp.solve().sat());
+		}
 	}
 	SECTION("test check mode") {
 		ClaspConfig config;
 		ClaspFacade libclasp;
 		class Prop : public Potassco::AbstractPropagator {
 		public:
-			Prop() : last(0) {}
+			Prop() : last(0), checks(0), props(0), totals(0), fire(false) {}
 			virtual void propagate(Potassco::AbstractSolver& s, const ChangeList& c) {
 				const Potassco::AbstractAssignment& a = s.assignment();
+				++props;
 				if (*Potassco::begin(c) == last) { return; }
 				for (int x = *Potassco::begin(c) + 1; a.hasLit(x); ++x) {
 					if (a.value(x) == Potassco::Value_t::Free) {
@@ -1543,30 +1742,356 @@ TEST_CASE("Clingo propagator", "[facade][propagator]") {
 			virtual void undo(const Potassco::AbstractSolver&, const ChangeList&) {}
 			virtual void check(Potassco::AbstractSolver& s) {
 				const Potassco::AbstractAssignment& a = s.assignment();
-				for (int x = 1; a.hasLit(x); ++x) {
-					if (a.value(x) == Potassco::Value_t::Free) {
-						s.addClause(Potassco::toSpan(&x, 1));
-						return;
+				++checks;
+				totals += a.isTotal();
+				if (fire) {
+					for (int x = 1; a.hasLit(x); ++x) {
+						if (a.value(x) == Potassco::Value_t::Free) {
+							s.addClause(Potassco::toSpan(&x, 1));
+							return;
+						}
 					}
+					REQUIRE(a.isTotal());
+					REQUIRE(a.level() == 0);
 				}
-				REQUIRE(a.isTotal());
-				REQUIRE(a.level() == 0);
 			}
 			int last;
+			int checks;
+			int props;
+			int totals;
+			bool fire;
 		} prop;
 		MyInit pp(prop);
 		config.addConfigurator(&pp);
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config);
 		lpAdd(asp, "{x1;x2;x3;x4;x5}.");
 		asp.endProgram();
-		pp.addWatch(posLit(1));
-		pp.addWatch(posLit(2));
-		pp.addWatch(posLit(3));
-		pp.addWatch(posLit(4));
-		pp.addWatch(posLit(5));
-		pp.enableClingoPropagatorCheck(ClingoPropagatorCheck_t::All);
-		libclasp.prepare();
-		REQUIRE(libclasp.ctx.master()->numFreeVars() == 0);
+		SECTION("test check and propagate") {
+			prop.fire = true;
+			pp.addWatch(posLit(1));
+			pp.addWatch(posLit(2));
+			pp.addWatch(posLit(3));
+			pp.addWatch(posLit(4));
+			pp.addWatch(posLit(5));
+			pp.enableClingoPropagatorCheck(ClingoPropagatorCheck_t::Fixpoint);
+			libclasp.prepare();
+			REQUIRE(libclasp.ctx.master()->numFreeVars() == 0);
+		}
+		SECTION("test check is called only once per fixpoint") {
+			pp.enableClingoPropagatorCheck(ClingoPropagatorCheck_t::Fixpoint);
+			libclasp.prepare();
+			REQUIRE(prop.checks == 1u);
+			libclasp.ctx.master()->propagate();
+			REQUIRE(prop.checks == 1u);
+			libclasp.ctx.master()->pushRoot(posLit(1));
+			REQUIRE(prop.checks == 2u);
+			libclasp.ctx.master()->assume(posLit(2)) && libclasp.ctx.master()->propagate();
+			REQUIRE(prop.checks == 3u);
+			libclasp.ctx.master()->propagate();
+			REQUIRE(prop.checks == 3u);
+			libclasp.ctx.master()->restart();
+			libclasp.ctx.master()->propagate();
+			INFO("Restart introduces new fix point");
+			REQUIRE(prop.checks == 4u);
+		}
+		SECTION("with mode total check is called once on total") {
+			pp.enableClingoPropagatorCheck(ClingoPropagatorCheck_t::Total);
+			libclasp.solve();
+			REQUIRE(prop.checks == 1u);
+			REQUIRE(prop.totals == 1u);
+		}
+		SECTION("with mode fixpoint check is called once on total") {
+			pp.enableClingoPropagatorCheck(ClingoPropagatorCheck_t::Fixpoint);
+			libclasp.solve();
+			REQUIRE(prop.checks > 1u);
+			REQUIRE(prop.totals == 1u);
+		}
 	}
 }
+
+TEST_CASE("Clingo propagator init", "[facade][propagator]") {
+	typedef ClingoPropagatorInit MyInit;
+	struct DebugLock : ClingoPropagatorLock {
+		DebugLock() : locked(false) {}
+		virtual void lock() {
+			REQUIRE(!locked);
+			locked = true;
+		}
+		virtual void unlock() {
+			REQUIRE(locked);
+			locked = false;
+		}
+		bool locked;
+	} debugLock;
+	PropagatorTest test;
+	SharedContext& ctx = test.ctx;
+	MyProp         prop;
+	MyInit         init(prop, &debugLock);
+
+	test.addVars(5);
+	init.prepare(ctx);
+	Solver& s0 = *ctx.master();
+	SECTION("add watches") {
+		init.addWatch(posLit(1));
+		init.addWatch(posLit(2));
+		init.addWatch(posLit(4));
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(s0.hasWatch(posLit(1), pp));
+		REQUIRE(s0.hasWatch(posLit(2), pp));
+		REQUIRE(s0.hasWatch(posLit(4), pp));
+		REQUIRE(!s0.hasWatch(posLit(3), pp));
+	}
+	SECTION("ignore duplicate watches from init") {
+		init.addWatch(posLit(1));
+		init.addWatch(posLit(1));
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(s0.hasWatch(posLit(1), pp));
+		s0.removeWatch(posLit(1), pp);
+		REQUIRE(!s0.hasWatch(posLit(1), pp));
+	}
+	SECTION("ignore duplicates on solver-specific init") {
+		init.addWatch(posLit(1));
+		init.addWatch(0, posLit(1));
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(s0.hasWatch(posLit(1), pp));
+		s0.removeWatch(posLit(1), pp);
+		REQUIRE(!s0.hasWatch(posLit(1), pp));
+	}
+	SECTION("add solver-specific watches") {
+		Solver& s1 = ctx.pushSolver();
+		init.prepare(ctx);
+		init.addWatch(posLit(1));     // add to both
+		init.addWatch(0, posLit(2));
+		init.addWatch(1, posLit(3));
+		init.addPost(s0);
+		init.addPost(s1);
+		ctx.endInit(true);
+		PostPropagator* pp0 = s0.getPost(PostPropagator::priority_class_general);
+		PostPropagator* pp1 = s1.getPost(PostPropagator::priority_class_general);
+		REQUIRE(s0.hasWatch(posLit(1), pp0));
+		REQUIRE(s1.hasWatch(posLit(1), pp1));
+
+		REQUIRE(s0.hasWatch(posLit(2), pp0));
+		REQUIRE(!s1.hasWatch(posLit(2), pp1));
+
+		REQUIRE(!s0.hasWatch(posLit(3), pp0));
+		REQUIRE(s1.hasWatch(posLit(3), pp1));
+	}
+	SECTION("don't add removed watch") {
+		Solver& s1 = ctx.pushSolver();
+		init.prepare(ctx);
+		// S0: [1,2,3]
+		// S1: [1, ,3]
+		init.addWatch(posLit(1));
+		init.addWatch(posLit(2));
+		init.addWatch(posLit(3));
+		init.removeWatch(1, posLit(2));
+		init.addPost(s0);
+		init.addPost(s1);
+		ctx.endInit(true);
+		PostPropagator* pp0 = s0.getPost(PostPropagator::priority_class_general);
+		PostPropagator* pp1 = s1.getPost(PostPropagator::priority_class_general);
+		REQUIRE(s0.hasWatch(posLit(1), pp0));
+		REQUIRE(s0.hasWatch(posLit(2), pp0));
+		REQUIRE(s0.hasWatch(posLit(3), pp0));
+
+		REQUIRE(s1.hasWatch(posLit(1), pp1));
+		REQUIRE(!s1.hasWatch(posLit(2), pp0));
+		REQUIRE(s1.hasWatch(posLit(3), pp1));
+	}
+
+	SECTION("last call wins") {
+		init.addWatch(posLit(1));
+		init.removeWatch(0, posLit(1));
+		init.addWatch(posLit(1));
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(s0.hasWatch(posLit(1), pp));
+	}
+
+	SECTION("watched facts are propagated") {
+		init.addWatch(posLit(1));
+		ctx.startAddConstraints();
+		ctx.addUnary(posLit(1));
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(prop.change.size() == 1);
+		REQUIRE(prop.change[0] == posLit(1));
+		REQUIRE(!s0.hasWatch(posLit(1), pp));
+	}
+	SECTION("facts can be watched even after propagate") {
+		init.addWatch(posLit(1));
+		ctx.startAddConstraints();
+		ctx.addUnary(posLit(1));
+		s0.propagate();
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(prop.change.size() == 1);
+		REQUIRE(prop.change[0] == posLit(1));
+		REQUIRE(!s0.hasWatch(posLit(1), pp));
+	}
+	SECTION("facts are propagated only once") {
+		init.addWatch(posLit(1));
+		ctx.startAddConstraints();
+		ctx.addUnary(posLit(1));
+		init.addPost(s0);
+		ctx.endInit();
+		PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+		REQUIRE(prop.change.size() == 1);
+		REQUIRE(prop.change[0] == posLit(1));
+		prop.change.clear();
+		ctx.unfreeze();
+		init.unfreeze(ctx);
+		ctx.startAddConstraints();
+		ctx.addUnary(posLit(2));
+		init.addWatch(posLit(1));
+		init.addWatch(posLit(2));
+		ctx.endInit();
+		REQUIRE(prop.change.size() == 1);
+		REQUIRE(prop.change[0] == posLit(2));
+	}
+
+	SECTION("init optionally keeps history so that future solvers get correct watches") {
+		init.enableHistory(true);
+		Solver& s1 = ctx.pushSolver();
+		init.prepare(ctx);
+		// S0: [1,2,3]
+		// S1: [1, ,3]
+		// S2: [ ,2, ,4]
+		init.addWatch(posLit(1));
+		init.addWatch(posLit(2));
+		init.addWatch(posLit(3));
+		init.removeWatch(1, posLit(2));
+		init.removeWatch(2, posLit(1));
+		init.removeWatch(2, posLit(3));
+		init.addWatch(2, posLit(4));
+		init.addPost(s0);
+		init.addPost(s1);
+		// don't add s2 yet
+		ctx.endInit(true);
+
+		ctx.unfreeze();
+		init.unfreeze(ctx);
+		Solver& s2 = ctx.pushSolver();
+		init.prepare(ctx);
+		ctx.startAddConstraints();
+		init.addWatch(posLit(5));
+		init.addPost(s2);
+		ctx.endInit(true);
+		PostPropagator* pp2 = s2.getPost(PostPropagator::priority_class_general);
+
+		REQUIRE(!s2.hasWatch(posLit(1), pp2));
+		REQUIRE(s2.hasWatch(posLit(2), pp2));
+		REQUIRE(!s2.hasWatch(posLit(3), pp2));
+		REQUIRE(s2.hasWatch(posLit(4), pp2));
+		REQUIRE(s2.hasWatch(posLit(5), pp2));
+	}
+
+	SECTION("test init-solve interplay") {
+		class Prop : public Potassco::AbstractPropagator {
+		public:
+			Prop() {}
+			virtual void propagate(Potassco::AbstractSolver& s, const ChangeList&) {}
+			virtual void undo(const Potassco::AbstractSolver&, const ChangeList&)  {}
+			virtual void check(Potassco::AbstractSolver& s) {
+				while (!add.empty()) {
+					s.addWatch(encodeLit(add.back()));
+					add.pop_back();
+				}
+				while (!remove.empty()) {
+					s.removeWatch(encodeLit(remove.back()));
+					remove.pop_back();
+				}
+			}
+			void addWatch(Literal lit)    { add.push_back(lit); }
+			void removeWatch(Literal lit) { remove.push_back(lit); }
+			LitVec add;
+			LitVec remove;
+		} prop;
+		MyInit init(prop);
+		init.enableClingoPropagatorCheck(ClingoPropagatorCheck_t::Fixpoint);
+
+		SECTION("ignore watches already added in init") {
+			init.addWatch(posLit(1));
+			prop.addWatch(posLit(1));
+			init.addPost(s0);
+			ctx.endInit();
+			PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+			REQUIRE(s0.hasWatch(posLit(1), pp));
+			REQUIRE(prop.add.empty());
+			s0.removeWatch(posLit(1), pp);
+			REQUIRE(!s0.hasWatch(posLit(1), pp));
+		}
+
+		SECTION("ignore watches in init already added during solving") {
+			prop.addWatch(posLit(1));
+			init.addPost(s0);
+			ctx.endInit();
+			PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+			REQUIRE(prop.add.empty());
+			ctx.unfreeze();
+			init.unfreeze(ctx);
+			ctx.startAddConstraints();
+			init.addWatch(posLit(1));
+			init.addWatch(posLit(2));
+			ctx.endInit();
+			REQUIRE(s0.hasWatch(posLit(1), pp));
+			REQUIRE(s0.hasWatch(posLit(2), pp));
+			s0.removeWatch(posLit(1), pp);
+			REQUIRE(!s0.hasWatch(posLit(1), pp));
+		}
+
+		SECTION("remove watch during solving") {
+			init.addWatch(posLit(1));
+			prop.removeWatch(posLit(1));
+			init.addPost(s0);
+			ctx.endInit();
+			PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+			ctx.unfreeze();
+			init.unfreeze(ctx);
+			ctx.startAddConstraints();
+			ctx.endInit();
+			REQUIRE(!s0.hasWatch(posLit(1), pp));
+		}
+
+		SECTION("remove watch during solving then add on init") {
+			init.addWatch(posLit(1));
+			prop.removeWatch(posLit(1));
+			init.addPost(s0);
+			ctx.endInit();
+			PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+			ctx.unfreeze();
+			init.unfreeze(ctx);
+			REQUIRE(!s0.hasWatch(posLit(1), pp));
+			init.addWatch(posLit(1));
+			ctx.startAddConstraints();
+			ctx.endInit();
+			REQUIRE(s0.hasWatch(posLit(1), pp));
+		}
+
+		SECTION("add watch during solving then remove on init") {
+			prop.addWatch(posLit(1));
+			init.addPost(s0);
+			ctx.endInit();
+			PostPropagator* pp = s0.getPost(PostPropagator::priority_class_general);
+			ctx.unfreeze();
+			init.unfreeze(ctx);
+			REQUIRE(s0.hasWatch(posLit(1), pp));
+			init.removeWatch(posLit(1));
+			ctx.startAddConstraints();
+			ctx.endInit();
+			REQUIRE(!s0.hasWatch(posLit(1), pp));
+		}
+	}
+}
+
 } }
