@@ -757,6 +757,14 @@ TEST_CASE("Logic program", "[asp]") {
 		REQUIRE(ctx.varInfo(lp.getLiteral(g).var()).frozen());
 	}
 
+	SECTION("testExternalsAreFrozen") {
+		lpAdd(lp.start(ctx),
+			"#external a.\n");
+		REQUIRE(lp.endProgram());
+		REQUIRE(lp.getLiteral(a).var() == 1);
+		REQUIRE(ctx.varInfo(lp.getLiteral(a).var()).frozen());
+	}
+
 	SECTION("testComputeTrueBug") {
 		lpAdd(lp.start(ctx),
 			"a :- not b.\n"
@@ -1204,15 +1212,18 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		lp.updateProgram();
 		lpAdd(lp,
 			"a.\n"
-			"b :- c.\n"
+			"b :- not c.\n"
 			"c.\n");
 		REQUIRE((lp.endProgram() && ctx.endInit()));
 		REQUIRE(ctx.numVars() == 0);
 		REQUIRE(lp.getLiteral(a) == lit_true());
-		REQUIRE(lp.getLiteral(b) == lit_true());
+		REQUIRE(lp.getLiteral(b) == lit_false());
+		REQUIRE(lp.getLiteral(b, MapLit_t::Refined) == lit_false());
 		REQUIRE(lp.getLiteral(c) == lit_true());
 		lp.updateProgram();
+		Var g = f + 1;
 		lpAdd(lp,
+			"g :- not f.\n"
 			"d.\n"
 			"e :- f.\n"
 			"f.\n");
@@ -1221,9 +1232,11 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		REQUIRE(lp.getLiteral(d) == lit_true());
 		REQUIRE(lp.getLiteral(e) == lit_true());
 		REQUIRE(lp.getLiteral(f) == lit_true());
+		REQUIRE(lp.getLiteral(g) == lit_false());
 		REQUIRE(lp.getLiteral(d, MapLit_t::Refined) == posLit(1));
 		REQUIRE(lp.getLiteral(e, MapLit_t::Refined) == posLit(1));
 		REQUIRE(lp.getLiteral(f, MapLit_t::Refined) == posLit(1));
+		REQUIRE(lp.getLiteral(g, MapLit_t::Refined) == negLit(1));
 	}
 	SECTION("testDistinctFactsSimple") {
 		lp.start(ctx, LogicProgram::AspOptions().noEq());
@@ -1499,6 +1512,37 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		REQUIRE(findSmodels(exp, lp));
 	}
 
+	SECTION("testSimplifyRuleEq") {
+		lp.start(ctx);
+		lp.updateProgram();
+		lpAdd(lp,
+			"{x3}.\n"
+			"x1 :- x3.\n"
+			"x2 :- x3.\n");
+
+		// x1 == x2
+		REQUIRE((lp.endProgram() && ctx.endInit()));
+		uint32 x1 = lp.getAtom(1)->id();
+		uint32 x2 = lp.getAtom(2)->id();
+		REQUIRE(x1 == x2);
+
+		lp.updateProgram();
+
+		lpAdd(lp,
+			"{x4}.\n"
+			"x5 :- 2{x1, x2, x4}.");
+		REQUIRE(lp.endProgram());
+		REQUIRE(lp.getAtom(5)->supports() == 1);
+		PrgBody* body = lp.getBody(lp.getAtom(5)->supps_begin()->node());
+		REQUIRE(body->type() == Potassco::Body_t::Sum);
+		REQUIRE(body->bound() == 2);
+		REQUIRE(body->size() == 2);
+
+		uint32 eqIdx = body->goal(0).var() == 4;
+		REQUIRE(body->weight(eqIdx) == 2);
+		REQUIRE(body->weight(1 - eqIdx) == 1);
+	}
+
 	SECTION("testEqUnfreeze") {
 		lp.start(ctx);
 		lp.updateProgram();
@@ -1951,6 +1995,32 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		}
 		REQUIRE((!foundA && foundB));
 	}
+
+	SECTION("testWriteExternalBug") {
+		lp.start(ctx);
+		lp.updateProgram();
+		lpAdd(lp,
+			"#external a."
+			"#external b."
+			"#external c."
+			"#external d."
+			"b.");
+		lp.endProgram();
+		std::stringstream str;
+		AspParser::write(lp, str, AspParser::format_aspif);
+		int foundA = 0, foundB = 0, foundC = 0, foundD = 0;
+		for (std::string x; std::getline(str, x);) {
+			if (x.find("5 1 2") == 0) { ++foundA; }
+			if (x.find("5 2 2") == 0) { ++foundB; }
+			if (x.find("5 3 2") == 0) { ++foundC; }
+			if (x.find("5 4 2") == 0) { ++foundD; }
+		}
+		REQUIRE(foundA == 1);
+		REQUIRE(foundB == 0);
+		REQUIRE(foundC == 1);
+		REQUIRE(foundD == 1);
+	}
+
 	SECTION("testWriteUnfreeze") {
 		lp.start(ctx);
 		lp.updateProgram();
@@ -2041,6 +2111,7 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		lp.endProgram();
 		REQUIRE_FALSE(lp.isExternal(a));
 	}
+
 	SECTION("testAssumptionsAreVolatile") {
 		lp.start(ctx);
 		lp.updateProgram();
@@ -2056,6 +2127,14 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		lp.getAssumptions(assume);
 		REQUIRE((assume.size() == 1 && assume[0] == ~lp.getLiteral(a)));
 	}
+
+	SECTION("testAssumptionsAreFrozen") {
+		lpAdd(lp.start(ctx), "{a}. #assume{a}.");
+		REQUIRE(lp.endProgram());
+		REQUIRE(lp.getLiteral(a).var() == 1);
+		REQUIRE(ctx.varInfo(lp.getLiteral(a).var()).frozen());
+	}
+
 	SECTION("testProjectionIsExplicitAndCumulative") {
 		lp.start(ctx);
 		lp.updateProgram();
